## ==========================================================================
## flowFrames are the basic data structures for flow cytometry data
## ==========================================================================






## ==========================================================================
## subsetting methods
## - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
## by indices or logical vectors
setMethod("[",
          signature=signature(x="flowFrame"),
          definition=function(x, i, j, ..., drop=FALSE)
      {
          if(drop)
              warning("Argument 'drop' ignored for subsetting of flowFrame")
          msg <- "Subset out of bounds"
          if(!missing(j)){
              if(is.logical(j))
                  if(max(which(j)) > ncol(x))
                      stop(msg, call.=FALSE)
              if(is.numeric(j))
                  if(max(abs(j)) > ncol(x))
                      stop(msg, call.=FALSE)
              if(is.character(j))
                  if(!all(j %in% colnames(x)))
                      stop(msg, call.=FALSE)
          }
          if(!missing(i))
               if(max(abs(i)) > nrow(x))
                   stop(msg, call.=FALSE)
          exprs(x) <-  switch(1+missing(i)+2*missing(j),
                          { exprs(x)[i, j, ..., drop=FALSE] },
                          { exprs(x)[ , j, ..., drop=FALSE] },
                          { exprs(x)[i,  , ..., drop=FALSE] },
                          { exprs(x)[ ,  , ..., drop=FALSE] } )
          
          x
      })

## by results of a filtering operation
setMethod("[",
          signature=signature(x="flowFrame",
                              i="filterResult"),
          definition=function(x,i,j,...,drop=FALSE)
      {
          if(missing(j))
              x[x %in% i,,...,drop=FALSE]
          else
              x[x %in% i,j,...,drop=FALSE]
      })

## by filter (which computes filter result first and applies that)
setMethod("[",
          signature=signature(x="flowFrame",
                              i="filter"),
          definition=function(x,i,j,...,drop=FALSE)
      {
          result <- filter(x,i)
          if(missing(j))
              x[result,,...,drop=FALSE]
          else
              x[result,j,...,drop=FALSE]
      })




## ==========================================================================
## the $-operator (subsetting with '$')
## - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
"$.flowFrame" <- function(x, val){
    if(!val %in% colnames(x))
        stop("'", val, "' is not a valid parameter in this flowFrame\n",
             "Subscript out of bounds")
    x[,val]
}




## ==========================================================================
## accessor and replace methods for slot exprs
## Note that only numeric matrices are accepted as replacement values
## and that the method checks if the colnames match the parameter
## definition in the parameters slot. Implicit subsetting is allowed
## (i.e. less columns in the replacement value compared to the original
## flowFrame, but all have to be defined there) 
## - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
setMethod("exprs",
          signature=signature(object="flowFrame"),
          definition=function(object){
              tmp <- object@exprs
              if(is(tmp, "ncdfHandler")){
                   if(!require(ncdf))
                       stop("You need to have package 'ncdf' installed in order for ",
                            "this feature to work.", call.=FALSE)
                  ## Need to figure out how to better use that
                  ##pointer <- if(!tmp@open) open.ncdf(tmp@file) else tmp@pointer
                  pointer <- open.ncdf(tmp@file)
                  tmp <- get.var.ncdf(pointer, "exprs")
                  if(!is.matrix(tmp)) 
                      tmp <- matrix(tmp, ncol=1)
                  colnames(tmp) <- object@parameters$name
                  close.ncdf(pointer)
              }
              return(tmp)
          })


## create a netCDF file from a data matrix for a single flowFrame
ncdfExpressionMatrix <- function(flowFrame,
                                 file=file.path(getwd(), ".flowCoreNcdf",
                                                paste(flowCore:::guid(),".ncdf", sep="")))
{
    if(nrow(flowFrame) == 0)
        return(NULL)
    if(!file.exists(".flowCoreNcdf"))
       dir.create(".flowCoreNcdf")
    ID <- "exprs"
    x <- dim.def.ncdf("cells", "counts", as.integer(seq_len(nrow(flowFrame))))
    y <- dim.def.ncdf("parameters", "stains", as.integer(seq_len(ncol(flowFrame))))
    dat <- var.def.ncdf(ID, "channels", list(x,y), NA, prec="double")
    if(is(flowFrame, "flowFrame")) flowFrame <- exprs(flowFrame)
    ncdf <- create.ncdf(file, dat)
    put.var.ncdf(ncdf, ID, flowFrame)
    close.ncdf(ncdf)
    return(list(file=file, pointer=ncdf))
}



setReplaceMethod("exprs",
                 signature=signature(object="flowFrame",
                                     value="matrix"),
                 definition=function(object, value)
             {
                 if(!is.numeric(value))
                     stop("replacement value for the 'exprs' slot ",
                          "of 'flowFrame' objects must be numeric",
                          "matrix", call.=FALSE)
                 mt <- match(colnames(value), colnames(object))
                 if(length(mt)==0)
                     stop("colnames missing in replacement value. ",
                          "Unable to match data columns", call.=FALSE)
                 if(any(is.na(mt)))
                     stop("the following parameters are not defined in the ",
                          "'flowFame' object:\n\t",
                          paste(colnames(value)[is.na(mt)], collapse=", "),
                          "\nunable to replace", call.=FALSE)
                 object@parameters <- object@parameters[mt,,drop=FALSE]
                 tmp <- object@exprs
                 if(is(tmp, "ncdfHandler")){
                     if(!require(ncdf))
                         stop("You need to have package 'ncdf' installed in order for ",
                              "this feature to work.", call.=FALSE)
                     ## The following overwrites files, but for now lets
                     ## create new ones.
                     ## res <- ncdfExpressionMatrix(value, tmp@file)
                     res <- ncdfExpressionMatrix(value)
                     if(!is.null(res)){
                         object@exprs@open <- TRUE
                         object@exprs@pointer <- res$pointer
                         object@exprs@file <- res$file
                     }else{
                         object@exprs <- value
                     }
                 }else{
                     object@exprs <- value
                 }
                 return(object)
             })


## throw meaningful error when trying to replace with anything other than matrix
setReplaceMethod("exprs",
                 signature=signature(object="flowFrame", value="ANY"),
                 definition=function(object, value)
                 stop("Replacement value for the 'exprs' slot of a ",
                      "'flowFrame' object must be \n  a numeric matrix with ",
                      "colnames matching at least a subset of the orginal",
                      "\n  columns."))



## ==========================================================================
## accessor and replace methods for slot description
## - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
setMethod("description",
          signature=signature(object="flowFrame"),
          function(object, hideInternal=FALSE){
              if(!hideInternal)
                  object@description
              else{
                  sel <- grep("^\\$", names(object@description))
                  object@description[-sel]
              }})

## replace description entries
descError <- "Replacement value must be a named list."
setReplaceMethod("description",
                 signature=signature(object="flowFrame",
                                     value="list"),
                 definition=function(object, value)
             {
                 n <- names(value)
                 if(length(n) == 0)
                     stop(descError, call.=FALSE)
                 object@description[n] <- value
                 return(object) })

setReplaceMethod("description",
                 signature=signature(object="flowFrame",
                                     value="ANY"),
                 definition=function(object, value)
                 stop(descError, call.=FALSE))


## ==========================================================================
## accessor methods for individual items in the description slot
## - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
## select keywords by name
setMethod("keyword",
          signature=signature(object="flowFrame",
                              keyword="character"),
          function(object, keyword)
              structure(object@description[keyword], names=keyword)
          )

## select or combine keywords by function
setMethod("keyword",
          signature=signature(object="flowFrame",
                              keyword="function"),
          definition=function(object,keyword)
          keyword(object)
          )

## select keywords by combination of name and/or function
setMethod("keyword",
          signature=signature(object="flowFrame",
                              keyword="list"),
          definition=function(object,keyword)
      {
          sapply(keyword, function(k) {
              if(is.character(k))
                  object@description[k]
              else if(is.function(k))
                  k(object)
              else NA
          })
      })

## this is equivalent to the description method
setMethod("keyword",
          signature=signature(object="flowFrame",
                              keyword="missing"),
          function(object)
              object@description
          )

## replace keywords
kwdError <- "Replacement value must be a named character vector or list."
setReplaceMethod("keyword",
                 signature=signature(object="flowFrame", value="character"),
                 definition=function(object, value)
             {
                 keyword(object) <- as.list(value)
                 return(object)
             })

setReplaceMethod("keyword",
                 signature=signature(object="flowFrame",
                                     value="list"),
                 definition=function(object, value)
             {
                 n <- names(value)
                 if(length(n) == 0)
                     stop(kwdError, call.=FALSE)
                 description(object) <- value
                 return(object)
             })

setReplaceMethod("keyword", signature=c("flowFrame", "ANY"),
                 definition=function(object, value)
                 stop(kwdError, call.=FALSE))



## ==========================================================================
## plot method: We actually need to attach flowViz to do the plotting
## - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
setMethod("plot",
          signature=signature(x="flowFrame",
                              y="ANY"),
          function(x, y, ...)
      {
          message("For plotting, please attach the 'flowViz' package.\n",
                  "   i.e., 'library(flowViz)'")
      })



## ==========================================================================
## nrow method
## - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
setMethod("nrow",
          signature=signature(x="flowFrame"),
          definition=function(x)
            return(nrow(exprs(x)))
          )



## ==========================================================================
## ncol method
## - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
setMethod("ncol",
          signature=signature(x="flowFrame"),
          definition=function(x)
            return(ncol(exprs(x)))
          )



## ==========================================================================
## dim method
## - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
setMethod("dim",
          signature=signature(x="flowFrame"),
          definition=function(x)
      {
          d <- dim(exprs(x))
          names(d) <- c("events", "parameters")
          return(d)
      })



## ==========================================================================
## accessor method for feature names
## - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
setMethod("featureNames",
          signature=signature(object="flowFrame"),
          definition=function(object)
          object@parameters$name
          )



## ==========================================================================
## accessor and replace methods for colnames of exprs slot
## this will also update the annotation in the parameters slot
## - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
setMethod("colnames",
          signature=signature(x="flowFrame"),
          definition=function(x, do.NULL="missing", prefix="missing")
          as.vector(x@parameters$name)
          )

setReplaceMethod("colnames",
                 signature=signature(x="flowFrame",
                                     value="ANY"),
                 definition=function(x, value)
             {
                 if(length(value) != ncol(x))
                     stop("colnames don't match dimensions of data matrix",
                          call.=FALSE)
                 pars <- parameters(x)
                 tmp <- exprs(x)
                 colnames(tmp) <- structure(as.character(value),
                                            names=rownames(pData(pars)))
                 pars$name <- value
                 names(pars$name) <- rownames(pData(pars))
                 x@parameters <- pars
                 exprs(x) <- tmp
                 return(x)
             })



## ===========================================================================
## compensate method
## ---------------------------------------------------------------------------
setMethod("compensate",
          signature=signature(x="flowFrame",
                              spillover="data.frame"),
          function(x, spillover)
      {
          compensate(x, as.matrix(spillover))
      })

setMethod("compensate",
          signature=signature(x="flowFrame",
                              spillover="matrix"),
          definition=function(x, spillover)
      {
#          checkClass(inv, "logical", 1)
#          if(inv)
#               spillover <- solve(spillover/max(spillover))
          cols <- colnames(spillover)
          sel <- cols %in% colnames(x)
          if(!all(sel))
              stop("The following parameters in the spillover matrix\n are",
                   " not present in the flowFrame:\n",
                   paste(cols[!sel], collapse=", "), call.=FALSE)
          e <- exprs(x)
		  ## pdh: this formula is now corrected
#          e[, cols] <- t(solve(spillover)%*%t(e[,cols]))
		  e[, cols] <- t(solve(t(spillover))%*%t(e[,cols]))
          exprs(x) = e
          x
      })

setMethod("compensate",
          signature=signature(x="flowFrame",
                              spillover="compensation"),
          function(x, spillover)
      {
          for(p in spillover@parameters)
              if(is(p, "transformReference"))
                  x  <- cbind2(x, matrix(resolveTransformReference(p, data),
                               dimnames=list(NULL, p@transformationId)))
          compensate(x, spillover@spillover)
      })



## ==========================================================================
## transform method
## we are also making sure that the values of the dynamic range in the
## parameters slot are transformed accordingly. Note that this does not
## happen in the inline form of transformation!
## - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
setMethod("transform",
          signature=signature(`_data`="flowFrame"),
          definition=function(`_data`, ...)
      {
          ## We hack method dispatch here to allow for something like
          ## transform(flowSet, transformList). Not really nice because
          ## we also need to deal with non-standard evaluation, but so what...
          suppressWarnings(try(if(length(list(...)) && is(..1, "transformList"))
              return(..1 %on% `_data`), silent=TRUE))
          e <- substitute(list(...))
          x <- `_data`
          par <- parameters(x)
          ranges <- range(x)
          tranges <- as.matrix(transform(as.data.frame(ranges),...))
          transformed <- as.matrix(transform(as.data.frame(exprs(x)),...))
          nc <- colnames(transformed)[-c(1:ncol(exprs(x)))]
          colnames(transformed) <- c(colnames(x), nc)
          if(ncol(transformed) > ncol(exprs(x))) {
              ## Add new parameter descriptions if there are any
              oCol <- lapply(e[2:length(e)], all.vars)
              nCol <- nc
              oparFrame <- pData(par)
              mt <- match(oCol[nc], oparFrame$name)
              nparFrame <- data.frame(name=nc,
                                      desc=paste("derived from",
                                                 "transformation of",
                                                 sapply(oCol[nc], paste,
                                                        collapse=" and ")),
                                      range=oparFrame[mt, "range"],
                                      minRange=tranges[1, nc],
                                      maxRange=tranges[2,nc])
              rownames(nparFrame) <- sprintf("$P%d", seq_len(nrow(nparFrame))+
                                             nrow(oparFrame))
              pData(par) <- rbind(oparFrame, nparFrame)
          }
          else {
                  cnames <- colnames(x)
                  par$minRange <- tranges[1,]
                  par$maxRange <- tranges[2,]
              }
          new("flowFrame",
              exprs=transformed, 
              parameters=par,
              description=description(x))
      })



## ==========================================================================
## filter method
## - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -


## apply filter
setMethod("filter",
          signature=signature(x="flowFrame",
                              filter="filter"),
          definition=function(x, filter)
      {
          temp <- resolveTransforms(x, filter)
          x <- temp[["data"]]
          filter <- temp[["filter"]]
          allPar <- parameters(filter) %in% colnames(x)
          if(!all(allPar))
              stop("The following parameter(s) are not present in this ",
                   "flowFrame:\n", paste("\t", parameters(filter)[!allPar],
                                         collapse="\n"), call.=FALSE)
          result <- as(x %in% filter, "filterResult")
          identifier(result) <- identifier(filter)
          filterDetails(result, identifier(filter)) <- filter
          result@frameId <- identifier(x)
          result
      })

## apply filterSet
setMethod("filter",
          signature=signature(x="flowFrame",
                              filter="filterSet"),
          definition=function(x,filter)
      {
        ## A list of filter names sorted by dependency
          fl <- sort(filter,dependencies=TRUE)
          e <- filterSet()
          m <- as(filter,"list")
          r <-  lapply(fl,function(n) {
              e[[n]] <- eval(m[[n]])
              filter(x,e[n])
          })
          ## A place to evaluate filters
          ## e  = new.env()
          ## assign(".__frame",x,env=e)
          ## Evaluate each 
          ## r = lapply(as(filter,"list")[fl],function(f) {
          ## if(is(f,"formula")) f = eval(as.call(c(as.symbol("as"),
          ##      f,"filter")),e)
          ##     print(f)
          ## eval(as.call(c(as.symbol("filter"),as.symbol(".__frame"),f)),e)	
          ## })
          if(all(sapply(r,is,"logicalFilterResult"))) {
              ## Combine into a many filter result
              manyFilterResult(r,frameId=identifier(x),attr(fl,'AdjM'))
          } else 
          r
      })



## ===========================================================================
## spillover method
## Flow frames can have a SPILL keyword that is often used to store the
## spillover matrix. Attempt to extract it.
## ---------------------------------------------------------------------------
setMethod("spillover",
          signature=signature(x="flowFrame"),
          definition=function(x)
      {
          present <- keyword(x, c("spillover", "SPILL"))
          if(all(sapply(present, is.null)))
              stop("No spillover matrix stored in that flowFrame")
          else present
      })



## ==========================================================================
## wrappers for apply (over exprs slot)
## - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
setMethod("each_row",
          signature=signature(x="flowFrame"),
          definition=function(x,FUN,...)
      {
          apply(exprs(x),1,FUN,...)
      })

setMethod("each_col",
          signature=signature(x="flowFrame"),
          definition=function(x,FUN,...)
      {
          apply(exprs(x),2,FUN,...)
      })



## ==========================================================================
## Subset methods for flowFrame
## Why do we need the 'select' parameter? Wouldn't this be equivalent:
## Subset(x[,c(1,3)], subset)
## - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -  
setMethod("Subset",
          signature=signature(x="flowFrame",
                              subset="filter"),
          definition=function(x,subset,select,...)
      {
          if(!missing(select))
              Subset(x,x %in% subset,select,...)
          else
              Subset(x,x %in% subset,...)
      })

setMethod("Subset",
          signature=signature(x="flowFrame",
                              subset="logical"),
          definition=function(x,subset,select,...)
      {
          if(!missing(select))
              x[subset & !is.na(subset),select]
          else x[subset,]
      })



## ==========================================================================
## range method for flowFrame
## get the range of possible data values from the parameters slot. Note that
## this range is not necessarily the range of the actual data values but
## it is the dynamic range of the measurement instrument
## - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
## Do we want the range to be set in a way to allow negative values?
## Need to fix that in read.FCS if yes
setMethod("range",
          signature=signature(x="flowFrame"),
          definition=function(x, ..., na.rm)
      {
          pind <- 1:nrow(parameters(x))
          channel <- unlist(list(...))
          if(length(channel)>0){
              if(!all(is(channel, "character")))
                  stop("All further arguments must be characters",
                       call.=FALSE)
                  pind <- match(channel, parameters(x)$name, nomatch=0)
          }
          if(any(pind==0))
              stop("'", paste(channel[which(pind==0)],
                              collapse="' and '", sep=""),
                   "' is/are no valid parameter(s) in this frame",
                   call.=FALSE) 
          ret <- rbind(min=parameters(x)$minRange[pind],
                       max=parameters(x)$maxRange[pind])
              colnames(ret) <- parameters(x)$name[pind]
          return(as.data.frame(ret))
      })



## ==========================================================================
## check for equality between two flowFrames
## - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
setMethod("==",
          signature=signature(e1="flowFrame",
                              e2="flowFrame"),
          definition=function(e1, e2)
      {
          return(all(as.numeric(exprs(e1)) == as.numeric(exprs(e2))) &&
                 all(as.character(pData(parameters(e1))) ==
                     as.character(pData(parameters(e2)))))
      })



## ==========================================================================
## show first or last values in the exprs matrix
## - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
setMethod("head",
          signature=signature(x="flowFrame"),
          definition=function(x, ...) head(exprs(x), ...))

setMethod("tail",
          signature=signature(x="flowFrame"),
          definition=function(x, ...) tail(exprs(x), ...))
          


## ==========================================================================
## comparison operators, these basically treat the flowFrame as a numeric
## matrix
## - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
setMethod("<",
          signature=signature(e1="flowFrame",
                              e2="ANY"),
          definition=function(e1, e2) exprs(e1) < e2)

setMethod(">",
          signature=signature(e1="flowFrame",
                              e2="ANY"),
          definition=function(e1, e2) exprs(e1) > e2)

setMethod("<=",
          signature=signature(e1="flowFrame",
                              e2="ANY"),
          definition=function(e1, e2) exprs(e1) <= e2)

setMethod(">=",
          signature=signature(e1="flowFrame",
                              e2="ANY"),
          definition=function(e1, e2) exprs(e1) >= e2)



## ==========================================================================
## add data columns to a flowFrame
## - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
setMethod("cbind2",
          signature=signature(x="flowFrame", y="matrix"),
          definition=function(x, y)
      {
          nx <- nrow(x)
          ny <- nrow(y)
          if(mode(y) != "numeric")
              stop("'y' must be numeric matrix.", call.=FALSE)
          cn <- colnames(y)
          if(is.null(cn) || any(cn %in% colnames(x)))
              stop("Invalid or missing colnames in 'y'", call.=FALSE)
          exp <- cbind(exprs(x), y)
          parms <- parameters(x)
          parm <- pData(parms)
          range <- t(apply(y, 2, range))
          colnames(range) <- c("minRange", "maxRange")
          parm <- rbind(parm, data.frame(name=cn, desc=NA, range=NA, range))
          pData(parms) <- parm
          x@parameters <- parms
          exprs(x) <- exp
          for(i in seq_along(cn)){
              tmp <- list(cn[i])
              names(tmp) <- sprintf("$P%dN", i+ncol(x))
              description(x) <- tmp
          }
          return(x)
      })

setMethod("cbind2",
          signature=signature(x="flowFrame", y="numeric"),
          definition=function(x, y)
          stop("'y' has to be numeric matrix with colnames.", call.=FALSE))
