## ==========================================================================
## Most objects in flowCore have a name or identifier, most prominently
## filters, flowFrames and flowSets. Please note that these identifiers
## don't necessarily have to be unique. Identifier methods provide a unified
## accessor to these IDs.
## ==========================================================================






## ==========================================================================
## Retrieve or set identifier of a flowFrame (normally generated by the
## cytometer and normally unique). There is a hierarchy of places in which
## to look for it: First, the "GUID" keyword, then the "FILENAME" and "$FIL"
## keywords, the fallback is to use the id "aninymous". Replacement will
## always fill the "GUID" keyword.
## - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
setMethod("identifier",
          signature=signature(object="flowFrame"),
          definition=function(object)
      {
          oid <- object@description[["GUID"]]
          if(is.null(oid) || is.na(oid))
              oid <- as.vector(object@description[["FILENAME"]])
          if(is.null(oid) || is.na(oid))
              oid <- as.vector(object@description[["$FIL"]])
          if(is.null(oid) || is.na(oid))
              "anonymous"
          else
              as.vector(oid)
      })

setReplaceMethod("identifier",
                 signature=signature(object="flowFrame",
                                     value="ANY"),
                 definition=function(object, value)
             {
                 object@description[["ORIGINALGUID"]] <- object@description[["GUID"]]
                 object@description[["GUID"]] <- value
                 tmp <- object@exprs
                 return(object)
             })



## ==========================================================================
## For compensation objects we can get the ID directly from the
## compensationId slot
## - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
setMethod("identifier",
          signature=signature(object="compensation"),
          definition=function(object) object@compensationId)

setReplaceMethod("identifier",
                 signature=signature("compensation","character"),
                 definition=function(object,value)
             {
                 object@compensationId <- value
                 object
             })



## ==========================================================================
## For actionItem objects we can get the ID directly from the ID slot
## - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
setMethod("identifier",
          signature=signature(object="actionItem"),
          definition=function(object) object@ID)



## ==========================================================================
## For fcReference objects we can get the ID directly from the ID slot
## - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
setMethod("identifier",
          signature=signature("fcReference"),
          definition=function(object) object@ID)


## ==========================================================================
## For filterList objects we can get the ID directly from the filterId slot
## - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
setMethod("identifier",
          signature=signature("filterList"),
          definition=function(object) object@filterId)

setReplaceMethod("identifier",
                 signature=signature("filterList","character"),
                 definition=function(object,value)
             {
                 object@filterId <- value
                 object
             })




## ==========================================================================
## For filterReferences we first resolve the reference to a concrete filter
## and use the next available method.
## - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
setMethod("identifier",
          signature=signature(object="filterReference"),
          definition=function(object)
      {
          if(exists(object@name, env=object@env))
              identifier(as(object, "concreteFilter"))
          else
              object@name
      })



## ==========================================================================
## For all filter objects we can get the ID directly from the filterId slot.
## The setter simply replaces this value.
## --------------------------------------------------------------------------
setMethod("identifier",
          signature=signature(object="filter"),
          definition=function(object) object@filterId)

setReplaceMethod("identifier",
                 signature=signature("filter","character"),
                 definition=function(object,value)
             {
                 object@filterId <- value
                 object
             })



## ==========================================================================
## For filterSet objects we can get the ID directly from the name slot
## --------------------------------------------------------------------------
setMethod("identifier",
          signature=signature(object="filterSet"),
          definition=function(object) object@name)



## ==========================================================================
## For view objects we can get the ID directly from the ID slot
## - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
setMethod("identifier",
          signature=signature(object="view"),
          definition=function(object) object@ID)



## ==========================================================================
## We need this to avoid type-checking
## - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
setMethod("identifier",
          signature=signature(object="NULL"),
          definition=function(object) NULL)


## ==========================================================================
## Transform objects have an ID slot
## - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
setMethod("identifier",
          signature=signature(object="transform"),
          definition=function(object) object@transformationId)



## ==========================================================================
## normalize objects have a normalizationId slot
## - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
setMethod("identifier",
          signature=signature(object="normalization"),
          definition=function(object) object@normalizationId)

setReplaceMethod("identifier",
                 signature=signature("normalization","character"),
                 definition=function(object,value)
             {
                 object@normalizationId <- value
                 object
             })


## ==========================================================================
## subsetting objects have a subsettingId slot
## - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
setMethod("identifier",
          signature=signature(object="subsetting"),
          definition=function(object) object@subsettingId)

setReplaceMethod("identifier",
                 signature=signature("subsetting","character"),
                 definition=function(object,value)
             {
                 object@subsettingId <- value
                 object
             })



## ==========================================================================
## transformList objects have a transformationId slot
## - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
setMethod("identifier",
          signature=signature(object="transformList"),
          definition=function(object) object@transformationId)

setReplaceMethod("identifier",
                 signature=signature("transformList","character"),
                 definition=function(object,value)
             {
                 object@transformationId <- value
                 object
             })
